package se.chalmers.paragon;

import java.util.HashSet;

/**
 * Special class for wrapping a list of actors. To be used in hashsets etc.
 * 
 * @author Paragon Developers (http://www.cse.chalmers.se/paragon)
 * 
 */
public class ActorList {

	/**
	 * The list of represented actors
	 */
	private Actor[] actors;

	/**
	 * Construct a new actor pair with the provided arguments
	 * 
	 * @param actors
	 *            Actors to create a pair for
	 */
	protected ActorList(Actor... actors) {
		this.actors = actors;
	}
	
	public static ActorList newActorList(Actor... actors) {
		return new ActorList(actors);
	}

	/**
	 * Returns the actor list represented by this pair
	 * 
	 * @return The actor list represented by this pair
	 */
	protected Actor[] getActors() {
		return this.actors;
	}

	/**
	 * 
	 * @return The length of the list of actors represented by this pair
	 */
	protected int arity() {
		return actors.length;
	}

	/**
	 * 
	 * @return True if all of the actors are instances of
	 *         {@link se.chalmers.paragon.ConcreteActor}, false otherwise
	 */
	public boolean isGround() {
		for (Actor a : actors)
			if (!(a instanceof ConcreteActor))
				return false;
		return true;
	}

	public boolean equals(Object _ap) {
		if (_ap == null)
			return false;
		if (_ap instanceof ActorList) {
			ActorList ap = (ActorList) _ap;
			if (this.arity() != ap.arity())
				return false;
			for (int i = 0; i < this.arity(); i++) {
				if (!this.getActors()[i].equals(ap.getActors()[i]))
					return false;
			}
			return true;
		}
		return false;
	}

	public int hashCode() {
		int hc = 0;
		for (Actor a : actors)
			hc += a.hashCode();
		return hc;
	}

	/**
	 * Assumes that all represented actors are an instance of the
	 * 'ConcreteActor' class
	 * 
	 * @return The represented Actor array, each element casted to a concrete
	 *         actor
	 */
	public ConcreteActor[] getActorsConcrete() {
		ConcreteActor[] res = new ConcreteActor[actors.length];
		for (int i = 0; i < actors.length; i++) {
			res[i] = (ConcreteActor) actors[i];
		}
		return res;
	}

	public String toString() {
		String res = "[";
		for (int i = 0; i < actors.length - 1; i++) {
			res += actors[i].toString() + ",";
		}
		if (actors.length > 0)
			res += actors[actors.length - 1].toString();
		res += "]";
		return res;
	}

	/**
	 * 
	 * @return The set of variables in this list (can be empty)
	 */
	public HashSet<ActorVariable> getVariables() {
		HashSet<ActorVariable> vars = new HashSet<ActorVariable>();
		for (Actor actor : actors)
			if (actor instanceof ActorVariable)
				vars.add((ActorVariable)actor);
		return vars;
	}

}
