{-# LANGUAGE DeriveDataTypeable #-}
module Language.Java.Paragon.TypeCheck.TypeMap where

import Language.Java.Paragon.Syntax
import Language.Java.Paragon.Pretty
import Language.Java.Paragon.Interaction

import Language.Java.Paragon.TypeCheck.Actors
import Language.Java.Paragon.TypeCheck.Policy
import Language.Java.Paragon.TypeCheck.Locks
import Language.Java.Paragon.TypeCheck.Types

import qualified Data.Map as Map

import Data.Generics.Uniplate.Data

import Data.Data


typeMapModule :: String
typeMapModule = typeCheckerBase ++ ".TypeMap"



data VarFieldSig = VSig { 
      varType   :: TcType,
      varPol    :: TcPolicy,
      varParam  :: Bool,
      varStatic :: Bool,
      varFinal  :: Bool
    }
  deriving (Show, Data, Typeable)

data MethodSig = MSig {
      mRetType :: TcType,
      mRetPol  :: TcPolicy,
      mPars    :: [TcPolicy],
      mWrites  :: TcPolicy,
      mExpects :: [TcLock],
      mLMods   :: ([TcLock],[TcLock]),
      mExns    :: [(TcType, ExnSig)]
    }
  deriving (Show, Data, Typeable)

data ExnSig = ExnSig { 
      exnReads :: TcPolicy, 
      exnWrites :: TcPolicy, 
      exnMods :: ([TcLock],[TcLock]) 
    }
  deriving (Show, Data, Typeable)

data ConstrSig = CSig {
      cPars    :: [TcPolicy],
      cWrites  :: TcPolicy,
      cExpects :: [TcLock],
      cLMods   :: ([TcLock],[TcLock]),
      cExns    :: [(TcType, ExnSig)]
    }
  deriving (Show, Data, Typeable)

data LockSig = LSig {
      lPol     :: TcPolicy,
      lArity   :: Int
    }
  deriving (Show, Data, Typeable)

data TypeSig = TSig {
      tType       :: TcType,
      tIsClass    :: Bool,
      tIsFinal    :: Bool,
      tSupers     :: [TcType],
      tImpls      :: [TcType],
      tMembers    :: TypeMap
    }
  deriving (Show, Data, Typeable)


type Map = Map.Map

data TypeMap = TypeMap {
      -- signatures
      fields      :: Map (Ident ()) VarFieldSig,
      methods     :: Map (Ident (), [TcType]) ([TypeParam ()], MethodSig),
      constrs     :: Map [TcType] ([TypeParam ()], ConstrSig),
      locks       :: Map (Ident ()) LockSig,
      -- known policy-level entities
      policies    :: Map (Ident ()) TcPolicy,
      actors      :: Map (Ident ()) ActorId,
      -- typemethod eval info
      typemethods :: Map (Ident ()) ([Ident ()], Block ()),
      -- types and packages
      types       :: Map (Ident ()) ([TypeParam ()], TypeSig),
      packages    :: Map (Ident ()) TypeMap
    }
  deriving (Show, Data, Typeable)

emptyTM :: TypeMap
emptyTM = TypeMap {
                fields      = Map.empty,
                methods     = Map.empty,
                constrs     = Map.empty,
                locks       = Map.empty,
                policies    = Map.empty,
                actors      = Map.empty,
                typemethods = Map.empty,
                types       = Map.empty,
                packages    = Map.empty
              }

hardCodedArrayTM :: TcType -> TcPolicy -> TypeMap
hardCodedArrayTM _ _ = 
    emptyTM {
         fields = Map.fromList [(Ident () "length", VSig intT thisP False False True)]
       , methods = Map.fromList [] -- TODO
       }

clearToPkgs :: TypeMap -> TypeMap
clearToPkgs tm = emptyTM { packages = packages tm }
                   
clearToTypes :: TypeMap -> TypeMap
clearToTypes tm = emptyTM { packages = packages tm, types = types tm }

pkgsAndTypes :: TypeMap -> Map (Ident ()) TypeMap
pkgsAndTypes tm = Map.union (packages tm)
                    -- disregard type parameters
                    (Map.map (tMembers . snd) $ types tm)


--------------------------------------------
--       Working with extensions          --
--------------------------------------------

extendTypeMapP :: Name () -> TypeMap -> TypeMap -> TypeMap
extendTypeMapP = go . flattenName
  where
    go :: [Ident ()] -> TypeMap -> TypeMap -> TypeMap
    go [] _ _ = panic (typeMapModule ++ ".extendTypeMapP") "Empty ident list"
    go [i] leafTm tm = 
        tm { packages = Map.insert i leafTm (packages tm) }
    go (i:is) leafTm tm =
        let mTm = packages tm 
            eTm = case Map.lookup i mTm of
                    Just innerTm -> innerTm
                    Nothing -> emptyTM
            newTm = go is leafTm eTm
        in tm { packages = Map.insert i newTm mTm }

extendTypeMapT :: Name () -> [TypeParam ()] -> TypeSig -> TypeMap -> TypeMap
extendTypeMapT = go . flattenName
  where
    go :: [Ident ()] -> [TypeParam ()] -> TypeSig -> TypeMap -> TypeMap
    go [] _ _ _ = panic (typeMapModule ++ ".extendTypeMapT")
                  "Empty ident list"
    go [i] tps tSig tm =
        tm { types = Map.insert i (tps, tSig) (types tm) }
    go (i:is) tps tSig tm =
        let mTm = packages tm 
            eTm = case Map.lookup i mTm of
                    Just innerTm -> innerTm
                    Nothing -> emptyTM
            newTm = go is tps tSig eTm
        in tm { packages = Map.insert i newTm mTm }

extendTypeMapN :: Name () -> (TypeMap -> TypeMap) -> TypeMap -> TypeMap
extendTypeMapN = go . flattenName
  where
    go :: [Ident ()] -> (TypeMap -> TypeMap) -> TypeMap -> TypeMap
    go [] _ _ = panic (typeMapModule ++ ".extendTypeMapN")
                "Empty ident list"
    go [i] tmf tm =
        let mTm = types tm
            (tps,tSig) = case Map.lookup i mTm of
                           Just tyInfo -> tyInfo
                           Nothing -> panic (typeMapModule ++ ".extendTypeMapN") $
                                      "Type not yet initialized: " ++ prettyPrint i
            tTm = tMembers tSig
            newSig = tSig { tMembers = tmf tTm }
        in tm { types = Map.insert i (tps, newSig) mTm }

    go (i:is) tmf tm =
        let mTm = packages tm
            eTm = case Map.lookup i mTm of
                    Just innerTm -> innerTm
                    Nothing -> panic (typeMapModule ++ ".extendTypeMapN") $
                               "Package not yet initialized: " ++ prettyPrint i
            newTm = go is tmf eTm
        in tm { packages = Map.insert i newTm mTm }

--------------------------------------
--    Working with the lookups      --
--------------------------------------

lookupNamed :: (TypeMap -> Map (Ident ()) a) -> Name () -> TypeMap -> Maybe a
lookupNamed recf (Name _ _ Nothing i) tm = Map.lookup i (recf tm)
lookupNamed recf nam@(Name _ _ (Just pre) i) tm = do
    newTm <- case nameType pre of
               TName -> do
                 (_tps, tsig) <- lookupNamed types pre tm
                 --if not (null tps)
                  --then Nothing
                  --else 
                 return $ tMembers tsig
               PName -> lookupNamed packages pre tm
{-               EName -> do
                 vsig <- lookupNamed fields pre tm
                 case lookupTypeOfT (varType vsig) tm of
                   Left _ -> Nothing
                   Right newTm -> return newTm -}
               _ -> panic (typeMapModule ++ ".lookupNamed") $
                    "Prefix is not a package or type: " ++ show nam
    Map.lookup i $ recf newTm
    
lookupNamed _ _ _ = panic (typeMapModule ++ ".lookupNamed")
                    "AntiQName should not appear in AST being type-checked"

-- | lookupTypeOfT will, given a type T and a top-level type environment,
--   return the type environment for T tagged with Right.
--   Left denotes an error, which wraps:
--   * If T is not a refType, return Nothing
--   * If T is given the wrong number of type arguments, return Just errorMessage.
lookupTypeOfT :: TcType -> TypeMap -> Either (Maybe String) TypeMap
lookupTypeOfT (TcRefT refT) = lookupTypeOfT' refT
lookupTypeOfT _ = const $ Left Nothing

lookupTypeOfT' :: TcRefType -> TypeMap -> Either (Maybe String) TypeMap
lookupTypeOfT' (TcArrayT ty pol) _ = Right $ hardCodedArrayTM ty pol
lookupTypeOfT' (TcClsRefT TcNullT) _ = Left $ Just $ "Cannot dereference null"
lookupTypeOfT' (TcTypeVar _ ) _ = panic (typeMapModule ++ ".lookupTypeOfT'")
                                  "TcTypeVar should have been instantiated"

lookupTypeOfT' (TcClsRefT (TcClassT n tas)) startTm =
    case n of
      Name _ TName _ _ -> 
          let mSig = lookupNamed types n startTm
          in case mSig of
               Nothing -> Left Nothing
               Just (tps, tsig) -> 
                   if length tps == length tas
                    then Right $ instantiate (zip tps tas) (tMembers tsig)
                    else Left $ Just $ 
                             "Wrong number of type arguments in class type.\n" ++
                             "Type " ++ prettyPrint n ++ " expects " ++ show (length tps) ++
                             " arguments but has been given " ++ show (length tas)
      Name _ _ _ _ -> Left Nothing
      _ -> panic (typeMapModule ++ ".lookupTypeOfT") $ show n

--------------------------------------
--   Type argument instantiation    --
--------------------------------------

instantiate :: Data a => [(TypeParam (),TcTypeArg)] -> a -> a
instantiate pas = transformBi instT
                     . transformBi instA
                     . transformBi instP
                     . transformBi instLs
    where instT :: TcRefType -> TcRefType
          instT tv@(TcTypeVar i) =
              case lookup i typs of
                Just rt -> rt
                Nothing -> tv
          instT rt = rt

          instA :: ActorId -> ActorId
          instA av@(ActorTPVar i) =
              case lookup i as of
                Just a -> a
                Nothing -> av
          instA a = a

          instP :: TcPolicy -> TcPolicy
          instP pv@(TcRigidVar i) =
              case lookup i ps of
                Just p -> p
                Nothing -> pv
          instP p = p

          instLs :: [TcLock] -> [TcLock]
          instLs = concatMap instL

          instL :: TcLock -> [TcLock]
          instL lv@(TcLockVar i) =
              case lookup i locs of
                Just le -> le
                Nothing -> [lv]
          instL l = [l]

          typs = [ (i, rt) | (TypeParam    _ i _, TcActualType      rt) <- pas ] 
          as   = [ (i, n ) | (ActorParam     _ i, TcActualActor      n) <- pas ]
          ps   = [ (i, p ) | (PolicyParam    _ i, TcActualPolicy     p) <- pas ]
          locs = [ (i, ls) | (LockStateParam _ i, TcActualLockState ls) <- pas ]
